#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <softhsm/pkcs11.h>

#define ARRAY_SIZE(arr) (sizeof(arr) / sizeof((arr)[0]))

#define SLOT_COUNT 128
#define LABEL_SIZE 32
#define RSA_BITS 2048
#define DATA_SIZE 64

/* the token must be configured with this label & PIN: */
static unsigned char LABEL[LABEL_SIZE] = "test                            ";
static unsigned char PIN[4] = "test";

static CK_BBOOL CTRUE  = TRUE;
static CK_BBOOL CFALSE = FALSE;

static CK_ULONG RSA_BITS_V = RSA_BITS;
static CK_BYTE RSA_PUB_EXP[] = { 0x01, 0x00, 0x01 };

static CK_ATTRIBUTE RSA_PUB_TEMPLATE[] = {
    { CKA_MODULUS_BITS,    &RSA_BITS_V,  sizeof(RSA_BITS_V)  },
    { CKA_ENCRYPT,         &CTRUE,       sizeof(CTRUE)       },
    { CKA_VERIFY,          &CTRUE,       sizeof(CTRUE)       },
    { CKA_PUBLIC_EXPONENT, RSA_PUB_EXP,  sizeof(RSA_PUB_EXP) }
};

static CK_ATTRIBUTE RSA_PRIV_TEMPLATE[] = {
    /* the private key shouldn't be extractable: */
    { CKA_EXTRACTABLE,  &CFALSE,    sizeof(CFALSE) },
    { CKA_DECRYPT,      &CTRUE,     sizeof(CTRUE)  },
    { CKA_SIGN,         &CTRUE,     sizeof(CTRUE)  }
};

static CK_OBJECT_CLASS PRIVATE_KEY = CKO_PRIVATE_KEY;
static CK_OBJECT_CLASS PUBLIC_KEY = CKO_PUBLIC_KEY;

static void dump_hex(const void *data, size_t size)
{
    const unsigned char *p = data;
    for (size_t i = 0; i < size; i++) {
        printf("%02x", (unsigned)p[i]);
    }
}

static int find_the_token(CK_SLOT_ID *slot)
{
    CK_RV ret;

    CK_SLOT_ID slots[SLOT_COUNT];

    /* get slot list: */
    unsigned long count = SLOT_COUNT;
    ret = C_GetSlotList(FALSE, slots, &count);
    if (ret != CKR_OK) {
        fprintf(stderr, "ERROR: C_GetSlotList(): %lu\n", ret);
        return 1;
    }

    printf("INFO: Found %lu slots!\n", count);

    /* find the desied token: */
    int token_found = 0;
    for (unsigned long i = 0; i < count; i++) {
        CK_TOKEN_INFO token;
        ret = C_GetTokenInfo(slots[i], &token);
        if (ret != CKR_OK) {
            fprintf(stderr, "ERROR: C_GetTokenInfo(): %lu\n", ret);
            return 1;
        }
        if (memcmp(LABEL, token.label, LABEL_SIZE) == 0) {
            token_found = 1;
            *slot = slots[i];
            break;
        }
    }
    if (!token_found) {
        fprintf(stderr, "ERROR: Token with label '%s' not found!\n", LABEL);
        return 1;
    }
    return 0;
}

static int generate_keypair(CK_SESSION_HANDLE session,
                            CK_OBJECT_HANDLE *pubkey, CK_OBJECT_HANDLE *privkey)
{
    CK_RV ret;

    CK_MECHANISM mech;
    mech.mechanism = CKM_RSA_PKCS_KEY_PAIR_GEN;
    mech.pParameter = NULL_PTR;
    mech.ulParameterLen = 0;

    ret = C_GenerateKeyPair(session, &mech,
                            RSA_PUB_TEMPLATE, ARRAY_SIZE(RSA_PUB_TEMPLATE),
                            RSA_PRIV_TEMPLATE, ARRAY_SIZE(RSA_PRIV_TEMPLATE),
                            pubkey, privkey);
    if (ret != CKR_OK) {
        fprintf(stderr, "ERROR: C_GenerateKeyPair(): %lu\n", ret);
        return 1;
    }
    return 0;
}

static int export_pubkey(CK_SESSION_HANDLE session, CK_OBJECT_HANDLE pubkey,
                         CK_BYTE *modulus, unsigned long *modulus_size,
                         CK_BYTE *exponent, unsigned long *exponent_size)
{
    CK_RV ret;

    CK_ATTRIBUTE templ[] = {
        { CKA_MODULUS, modulus, *modulus_size },
        { CKA_PUBLIC_EXPONENT, exponent, *exponent_size },
    };

    ret = C_GetAttributeValue(session, pubkey, templ, ARRAY_SIZE(templ));
    if (ret != CKR_OK) {
        fprintf(stderr, "ERROR: C_GetAttributeValue(): %lu\n", ret);
        return 1;
    }

    *modulus_size = templ[0].ulValueLen;
    *exponent_size = templ[1].ulValueLen;
    return 0;
}

static int find_keys(CK_SESSION_HANDLE session,
                     CK_BYTE *modulus, unsigned long modulus_size,
                     CK_BYTE *exponent, unsigned long exponent_size,
                     CK_OBJECT_HANDLE *pubkey, CK_OBJECT_HANDLE *privkey)
{
    CK_RV ret;

    /* find the public key: */
    CK_ATTRIBUTE templ[] = {
        { CKA_CLASS, &PUBLIC_KEY, sizeof(PUBLIC_KEY) },
        { CKA_MODULUS, modulus, modulus_size },
        { CKA_PUBLIC_EXPONENT, exponent, exponent_size }
    };

    ret = C_FindObjectsInit(session, templ, ARRAY_SIZE(templ));
    if (ret != CKR_OK) {
        fprintf(stderr, "ERROR: C_FindObjectsInit(): %lu\n", ret);
        return 1;
    }

    unsigned long count;
    ret = C_FindObjects(session, pubkey, 1, &count);
    if (ret != CKR_OK) {
        fprintf(stderr, "ERROR: C_FindObjects(): %lu\n", ret);
        return 1;
    }

    if (count < 1) {
        fprintf(stderr, "ERROR: Private key not found!\n");
        return 1;
    }

    ret = C_FindObjectsFinal(session);
    if (ret != CKR_OK) {
        fprintf(stderr, "ERROR: C_FindObjectsFinal(): %lu\n", ret);
        return 1;
    }

    /* find the private key: */
    templ[0].pValue = &PRIVATE_KEY;
    templ[0].ulValueLen = sizeof(PRIVATE_KEY);

    ret = C_FindObjectsInit(session, templ, ARRAY_SIZE(templ));
    if (ret != CKR_OK) {
        fprintf(stderr, "ERROR: C_FindObjectsInit(): %lu\n", ret);
        return 1;
    }

    ret = C_FindObjects(session, privkey, 1, &count);
    if (ret != CKR_OK) {
        fprintf(stderr, "ERROR: C_FindObjects(): %lu\n", ret);
        return 1;
    }

    if (count < 1) {
        fprintf(stderr, "ERROR: Private key not found!\n");
        return 1;
    }

    ret = C_FindObjectsFinal(session);
    if (ret != CKR_OK) {
        fprintf(stderr, "ERROR: C_FindObjectsFinal(): %lu\n", ret);
        return 1;
    }
    return 0;
}

static int encrypt_data(CK_SESSION_HANDLE session, CK_OBJECT_HANDLE pubkey,
                        CK_BYTE *data_in, unsigned long data_in_size,
                        CK_BYTE *data_out, unsigned long *data_out_size)
{
    CK_RV ret;

    CK_MECHANISM mech = { CKM_RSA_PKCS, NULL_PTR, 0 };
    ret = C_EncryptInit(session, &mech, pubkey);
    if (ret != CKR_OK) {
        fprintf(stderr, "ERROR: C_EncryptInit(): %lu\n", ret);
        return 1;
    }

    ret = C_Encrypt(session, data_in, data_in_size, data_out, data_out_size);
    if (ret != CKR_OK) {
        fprintf(stderr, "ERROR: C_Encrypt(): %lu\n", ret);
        return 1;
    }
    return 0;
}

static int decrypt_data(CK_SESSION_HANDLE session, CK_OBJECT_HANDLE privkey,
                        CK_BYTE *data_in, unsigned long data_in_size,
                        CK_BYTE *data_out, unsigned long *data_out_size)
{
    CK_RV ret;

    CK_MECHANISM mech = { CKM_RSA_PKCS, NULL_PTR, 0 };
    ret = C_DecryptInit(session, &mech, privkey);
    if (ret != CKR_OK) {
        fprintf(stderr, "ERROR: C_DecryptInit(): %lu\n", ret);
        return 1;
    }

    ret = C_Decrypt(session, data_in, data_in_size, data_out, data_out_size);
    if (ret != CKR_OK) {
        fprintf(stderr, "ERROR: C_Decrypt(): %lu\n", ret);
        return 1;
    }
    return 0;
}

static int destroy_keys(CK_SESSION_HANDLE session, CK_OBJECT_HANDLE pubkey,
                        CK_OBJECT_HANDLE privkey)
{
    CK_RV ret;
    ret = C_DestroyObject(session, pubkey);
    if (ret != CKR_OK) {
        fprintf(stderr, "ERROR: C_DestroyObject(): %lu\n", ret);
        return 1;
    }
    ret = C_DestroyObject(session, privkey);
    if (ret != CKR_OK) {
        fprintf(stderr, "ERROR: C_DestroyObject(): %lu\n", ret);
        return 1;
    }
    return 0;
}

int main(void)
{
    CK_RV ret;

    ret = C_Initialize(NULL);
    if (ret != CKR_OK) {
        fprintf(stderr, "ERROR: C_Initialize(): %lu\n", ret);
        return 1;
    }

    /* find the desired token: */
    printf("INFO: Looking for a token with label '%.*s'...\n", LABEL_SIZE, LABEL);
    CK_SLOT_ID slot;
    if (find_the_token(&slot)) {
        return 1;
    }

    /* open session: */
    CK_SESSION_HANDLE session;
    ret = C_OpenSession(slot, CKF_RW_SESSION | CKF_SERIAL_SESSION,
                        NULL_PTR, NULL_PTR, &session);
    if (ret != CKR_OK) {
        fprintf(stderr, "ERROR: C_OpenSession(): %lu\n", ret);
        return 1;
    }

    /* try generating the key pair without logging in: */
    printf("INFO: Trying to do something without logging in...\n");
    CK_OBJECT_HANDLE pubkey, privkey;
    if (!generate_keypair(session, &pubkey, &privkey)) {
        fprintf(stderr, "ERROR: The operation should have failed, but didn't!\n");
        return 1;
    }
    printf("INFO: Nope! let's try to login now...\n");

    /* login as a user: */
    ret = C_Login(session, CKU_USER, PIN, ARRAY_SIZE(PIN));
    if (ret != CKR_OK) {
        fprintf(stderr, "ERROR: C_Login(): %lu\n", ret);
        return 1;
    }

    /* generate the key pair: */
    printf("INFO: OK, now it should work...\n");
    if (generate_keypair(session, &pubkey, &privkey)) {
        return 1;
    }
    printf("INFO: Yeah, the keypair was generated successfully!\n");

    /* export public key: */
    CK_BYTE modulus[RSA_BITS / 8];
    CK_BYTE exponent[RSA_BITS / 8];
    unsigned long modulus_size = sizeof(modulus);
    unsigned long exponent_size = sizeof(exponent);
    if (export_pubkey(session, pubkey, modulus, &modulus_size,
                      exponent, &exponent_size)) {
        return 1;
    }

    /* imagine this is written into a file...: */
    printf("INFO: Modulus:  "); dump_hex(modulus, modulus_size); printf("\n");
    printf("INFO: Exponent: "); dump_hex(exponent, exponent_size); printf("\n");

    /* find the keys based on the exported modulus and exponent: */
    pubkey = -1;
    privkey = -1;
    if (find_keys(session, modulus, modulus_size,
                  exponent, exponent_size, &pubkey, &privkey)) {
        return 1;
    }

    CK_BYTE data[RSA_BITS / 8];
    unsigned long data_size = sizeof(data);

    for (size_t i = 0; i < DATA_SIZE; i++) {
        data[i] = (CK_BYTE)rand();
    }

    printf("INFO: Now encrypting "); dump_hex(data, DATA_SIZE); printf("...\n");

    /* encrypt some data: */
    if (encrypt_data(session, pubkey, data, DATA_SIZE, data, &data_size)) {
        return 1;
    }
    printf("INFO: Encrypted data: "); dump_hex(data, data_size); printf("\n");

    /* decrypt the data back: */
    printf("INFO: Now decrypting the data back...\n");
    if (decrypt_data(session, privkey, data, data_size, data, &data_size)) {
        return 1;
    }
    printf("INFO: Decrypted data: "); dump_hex(data, data_size); printf("\n");

    /* destroy the keys: */
    printf("INFO: Destroying the keys...\n");
    if (destroy_keys(session, pubkey, privkey)) {
        return 1;
    }

    /* close the session: */
    ret = C_CloseSession(session);
    if (ret != CKR_OK) {
        fprintf(stderr, "ERROR: C_CloseSession(): %lu\n", ret);
        return 1;
    }

    return 0;
}

